from __future__ import unicode_literals

import django

if django.VERSION[:2] < (1, 6):     # unittest-style discovery isn't available
    from .test_get_setting import SettingsTestCase
    from .test_http_server import HttpServerTestCase
    from .test_regression import RegressionTestCase
    from .test_storage import DistributedStorageTestCase
    from .test_storage import HybridStorageTestCase
    from .test_storage import AsyncStorageTestCase
    from .test_storage import DistributedStorageWithTwoServersTestCase
    from .test_storage import HybridStorageWithTwoServersTestCase
    from .test_storage import AsyncStorageWithTwoServersTestCase
    from .test_storage import DistributedStorageMiscTestCase
    from .test_storage import HybridStorageMiscTestCase
    from .test_storage import AsyncStorageMiscTestCase



# coding: utf-8

from __future__ import unicode_literals

import datetime
import io
import logging
import os
import os.path
import pickle
import shutil
import threading
import time
try:
    from urllib.request import HTTPError
except ImportError:
    from urllib2 import HTTPError

from django.conf import settings
from django.core.files.base import ContentFile
from django.utils import unittest

from .storage import (DistributedStorage, HybridStorage, AsyncStorage,
        UnexpectedStatusCode)
from .test_http_server import HttpServerTestCaseMixin, ExtraHttpServerTestCaseMixin


CONTENT = '¡test!'.encode('utf-8')


class StorageUtilitiesMixin(HttpServerTestCaseMixin):

    def setUp(self):
        super(StorageUtilitiesMixin, self).setUp()
        self.log = io.StringIO()
        self.logger = logging.getLogger('django_resto.storage')
        self.handler = logging.StreamHandler(self.log)
        self.logger.addHandler(self.handler)
        hosts = ['%s:%d' % (self.host, self.port)]
        self.storage = self.storage_class(hosts=hosts)
        if self.use_fs:
            os.makedirs(settings.MEDIA_ROOT)

    def tearDown(self):
        super(StorageUtilitiesMixin, self).tearDown()
        if self.use_fs:
            shutil.rmtree(settings.MEDIA_ROOT)

    def sync(self):
        pass

    def get_log(self):
        self.sync()
        self.handler.flush()
        return self.log.getvalue()

    def has_file(self, name):
        self.sync()
        return self.http_server.has_file(name)

    def get_file(self, name):
        self.sync()
        return self.http_server.get_file(name)

    def create_file(self, name, content):
        self.http_server.create_file(name, content)
        if self.use_fs:
            filename = os.path.join(settings.MEDIA_ROOT, name)
            if not os.path.isdir(os.path.dirname(filename)):
                os.makedirs(os.path.dirname(filename))
            with open(filename, 'wb') as f:
                f.write(content)

    def delete_file(self, name):
        self.http_server.delete_file(name)
        if self.use_fs:
            filename = os.path.join(settings.MEDIA_ROOT, name)
            os.unlink(filename)


class StorageUtilitiesWithTwoServersMixin(
        ExtraHttpServerTestCaseMixin, StorageUtilitiesMixin):

    def setUp(self):
        super(StorageUtilitiesWithTwoServersMixin, self).setUp()
        hosts = ['%s:%d' % (self.host, self.port + 1 - i) for i in range(2)]
        self.storage = self.storage_class(hosts=hosts)

    def create_file(self, name, content):
        self.alt_http_server.create_file(name, content)
        super(StorageUtilitiesWithTwoServersMixin, self).create_file(name, content)

    def delete_file(self, name):
        self.alt_http_server.delete_file(name)
        super(StorageUtilitiesWithTwoServersMixin, self).delete_file(name)


class HybridStorageTestCaseMixin(object):

    # Tests for HybridStorage, see variants for DistributedStorage below

    # See http://docs.djangoproject.com/en/dev/ref/files/storage/
    # for the full storage API.
    # See http://docs.djangoproject.com/en/dev/howto/custom-file-storage/
    # for a list of which methods a custom storage class must implement.

    def test_accessed_time(self):
        self.create_file('test.txt', CONTENT)
        self.assertIsInstance(self.storage.accessed_time('test.txt'), datetime.datetime)
        self.assertEachServerLogIs([])  # overridden when self.use_fs = False

    def test_accessed_time_non_existing(self):
        with self.assertRaises(EnvironmentError):
            self.storage.accessed_time('test.txt')
        self.assertEachServerLogIs([])  # overridden when self.use_fs = False

    def test_created_time(self):
        self.create_file('test.txt', CONTENT)
        self.assertIsInstance(self.storage.created_time('test.txt'), datetime.datetime)
        self.assertEachServerLogIs([])  # overridden when self.use_fs = False

    def test_created_time_non_existing(self):
        with self.assertRaises(EnvironmentError):
            self.storage.created_time('test.txt')
        self.assertEachServerLogIs([])  # overridden when self.use_fs = False

    def test_delete(self):
        self.create_file('test.txt', CONTENT)
        self.assertTrue(self.has_file('test.txt'))
        self.storage.delete('test.txt')
        self.assertFalse(self.has_file('test.txt'))
        self.assertEachServerLogIs([('DELETE', '/test.txt', 204)])

    def test_delete_non_existing(self):
        # by design, this doesn't raise an exception, but it logs a warning
        self.storage.delete('test.txt')
        self.assertFalse(self.has_file('test.txt'))
        self.assertIn("DELETE on missing file", self.get_log())
        self.assertEachServerLogIs([('DELETE', '/test.txt', 404)])

    def test_delete_readonly(self):
        self.create_file('test.txt', CONTENT)
        self.http_server.readonly = True
        with self.assertRaises(HTTPError):
            self.storage.delete('test.txt')
        self.assertIn("Failed to delete", self.get_log())
        self.assertServerLogIs([('DELETE', '/test.txt', 403)])

    def test_delete_dilettante(self):
        self.http_server.override_code = 202
        self.create_file('test.txt', CONTENT)
        with self.assertRaises(UnexpectedStatusCode):
            self.storage.delete('test.txt')
        self.assertIn("Failed to delete", self.get_log())
        self.assertServerLogIs([('DELETE', '/test.txt', 202)])

    def test_exists(self):
        self.create_file('test.txt', CONTENT)
        self.assertTrue(self.storage.exists('test.txt'))
        self.delete_file('test.txt')
        self.assertFalse(self.storage.exists('test.txt'))
        if self.use_fs:
            self.assertAnyServerLogIs([])
        else:
            self.assertAnyServerLogIs([('HEAD', '/test.txt', 200),
                                       ('HEAD', '/test.txt', 404)])

    def test_get_available_name(self):
        self.assertEqual(self.storage.get_available_name('test.txt'), 'test.txt')
        if self.use_fs:
            self.assertAnyServerLogIs([])
        else:
            self.assertAnyServerLogIs([('HEAD', '/test.txt', 404)])

    def test_get_available_name_existing(self):
        self.create_file('test.txt', CONTENT)
        self.assertEqual(self.storage.get_available_name('test.txt'), 'test_1.txt')
        if self.use_fs:
            self.assertAnyServerLogIs([])
        else:
            self.assertAnyServerLogIs([('HEAD', '/test.txt', 200),
                                       ('HEAD', '/test_1.txt', 404)])

    def test_get_valid_name(self):
        self.assertEqual(self.storage.get_valid_name('test.txt'), 'test.txt')
        self.assertEachServerLogIs([])

    def test_listdir(self):
        self.create_file('test/foo.txt', b'foo')
        self.create_file('test/bar.txt', b'bar')
        self.create_file('test/baz/quux.txt', b'quux')
        listing = self.storage.listdir('test')
        self.assertEqual(set(listing[0]), set(['baz']))
        self.assertEqual(set(listing[1]), set(['foo.txt', 'bar.txt']))
        self.assertEachServerLogIs([])  # overridden when self.use_fs = False

    def test_listdir_non_existing(self):
        with self.assertRaises(EnvironmentError):
            self.storage.listdir('test')
        self.assertEachServerLogIs([])  # overridden when self.use_fs = False

    def test_modified_time(self):
        self.create_file('test.txt', CONTENT)
        self.assertIsInstance(self.storage.modified_time('test.txt'), datetime.datetime)
        self.assertEachServerLogIs([])  # overridden when self.use_fs = False

    def test_modified_time_non_existing(self):
        with self.assertRaises(EnvironmentError):
            self.storage.modified_time('test.txt')
        self.assertEachServerLogIs([])  # overridden when self.use_fs = False

    def test_open(self):
        self.create_file('test.txt', CONTENT)
        with self.storage.open('test.txt') as f:
            self.assertEqual(f.read(), CONTENT)
        if self.use_fs:
            self.assertAnyServerLogIs([])
        else:
            self.assertAnyServerLogIs([('GET', '/test.txt', 200)])

    def test_path(self):
        self.create_file('test.txt', CONTENT)
        self.assertEqual(self.storage.path('test.txt'),
                         os.path.join(settings.MEDIA_ROOT, 'test.txt'))
        self.assertEachServerLogIs([])  # overridden when self.use_fs = False

    def test_save(self):
        filename = self.storage.save('test.txt', ContentFile(CONTENT))
        self.assertEqual(filename, 'test.txt')
        self.assertEqual(self.get_file('test.txt'), CONTENT)
        if self.use_fs:
            self.assertEachServerLogIs([('PUT', '/test.txt', 201)])
        else:
            self.assertAnyServerLogIs([('HEAD', '/test.txt', 404)] +
                                      [('PUT', '/test.txt', 201)] * len(self.storage.hosts))

    def test_save_existing(self):
        self.create_file('test.txt', CONTENT)
        filename = self.storage.save('test.txt', ContentFile(CONTENT * 2))
        # a new file is generated
        self.assertEqual(filename, 'test_1.txt')
        self.assertEqual(self.get_file('test_1.txt'), CONTENT * 2)
        if self.use_fs:
            self.assertEachServerLogIs([('PUT', '/test_1.txt', 201)])
        else:
            self.assertAnyServerLogIs([('HEAD', '/test.txt', 200),
                                       ('HEAD', '/test_1.txt', 404)] +
                                      [('PUT', '/test_1.txt', 201)] * len(self.storage.hosts))

    def test_save_readonly(self):
        self.http_server.readonly = True
        with self.assertRaises(HTTPError):
            self.storage.save('test.txt', ContentFile(CONTENT))
        action = 'upload' if self.use_fs else 'create'
        self.assertIn("Failed to %s" % action, self.get_log())
        if self.use_fs:
            self.assertServerLogIs([('PUT', '/test.txt', 403)])
        else:
            self.assertAnyServerLogIs([('HEAD', '/test.txt', 404),
                                       ('PUT', '/test.txt', 403)] +
                                      [('PUT', '/test.txt', 201)] * (len(self.storage.hosts) - 1))

    def test_save_dilettante(self):
        self.http_server.override_code = 202
        with self.assertRaises(UnexpectedStatusCode):
            self.storage.save('test.txt', ContentFile(CONTENT))
        action = 'upload' if self.use_fs else 'create'
        self.assertIn("Failed to %s" % action, self.get_log())
        # overridden in DistributedStorageTestCaseMixin
        self.assertServerLogIs([('PUT', '/test.txt', 202)])

    def test_size(self):
        self.create_file('test.txt', CONTENT)
        self.assertEqual(self.storage.size('test.txt'), 7)
        if self.use_fs:
            self.assertAnyServerLogIs([])
        else:
            self.assertAnyServerLogIs([('HEAD', '/test.txt', 200)])

    def test_url(self):
        self.create_file('test.txt', CONTENT)
        self.assertEqual(self.storage.url('test.txt'),
                'http://media.example.com/test.txt')
        self.assertEachServerLogIs([])


class AsyncStorageTestCaseMixin(HybridStorageTestCaseMixin):

    # Adapt for DistributedStorage tests of methods that behave differently.

    def test_delete_readonly(self):
        self.create_file('test.txt', CONTENT)
        self.http_server.readonly = True
        # Exceptions aren't checked
        self.storage.delete('test.txt')
        self.assertIn("Failed to delete", self.get_log())
        self.assertServerLogIs([('DELETE', '/test.txt', 403)])

    def test_delete_dilettante(self):
        self.http_server.override_code = 202
        self.create_file('test.txt', CONTENT)
        # Exceptions aren't checked
        self.storage.delete('test.txt')
        self.assertIn("Failed to delete", self.get_log())
        self.assertServerLogIs([('DELETE', '/test.txt', 202)])

    def test_save_readonly(self):
        self.http_server.readonly = True
        # Exceptions aren't checked
        self.storage.save('test.txt', ContentFile(CONTENT))
        self.assertIn("Failed to upload", self.get_log())
        self.assertServerLogIs([('PUT', '/test.txt', 403)])

    def test_save_dilettante(self):
        self.http_server.override_code = 202
        # Exceptions aren't checked
        self.storage.save('test.txt', ContentFile(CONTENT))
        self.assertIn("Failed to upload", self.get_log())
        self.assertServerLogIs([('PUT', '/test.txt', 202)])


class DistributedStorageTestCaseMixin(HybridStorageTestCaseMixin):

    # Adapt for DistributedStorage tests of methods that behave differently,
    # and disable tests that aren't deterministic with an inconsistent state.

    def test_accessed_time(self):
        with self.assertRaises(NotImplementedError):
            self.storage.accessed_time('test.txt')

    test_accessed_time_non_existing = test_accessed_time

    def test_created_time(self):
        with self.assertRaises(NotImplementedError):
            self.storage.created_time('test.txt')

    test_created_time_non_existing = test_created_time

    def test_exists_broken(self):
        if len(self.storage.hosts) > 1:
            return
        self.http_server.override_code = 500
        with self.assertRaises(HTTPError):
            self.storage.exists('test.txt')
        self.assertIn("Failed to check", self.get_log())
        self.assertServerLogIs([('HEAD', '/test.txt', 500)])

    def test_exists_dilettante(self):
        if len(self.storage.hosts) > 1:
            return
        self.http_server.override_code = 202
        with self.assertRaises(UnexpectedStatusCode):
            self.storage.exists('test.txt')
        self.assertIn("Failed to check", self.get_log())
        self.assertServerLogIs([('HEAD', '/test.txt', 202)])

    def test_listdir(self):
        with self.assertRaises(NotImplementedError):
            self.storage.listdir('test')

    test_listdir_non_existing = test_listdir

    def test_modified_time(self):
        with self.assertRaises(NotImplementedError):
            self.storage.modified_time('test.txt')

    test_modified_time_non_existing = test_modified_time

    def test_open_broken(self):
        if len(self.storage.hosts) > 1:
            return
        self.http_server.override_code = 500
        with self.assertRaises(HTTPError):
            self.storage.open('test.txt')
        self.assertIn("Failed to download", self.get_log())
        self.assertAnyServerLogIs([('GET', '/test.txt', 500)])

    def test_open_dilettante(self):
        if len(self.storage.hosts) > 1:
            return
        self.http_server.override_code = 202
        with self.assertRaises(UnexpectedStatusCode):
            self.storage.open('test.txt')
        self.assertIn("Failed to download", self.get_log())
        self.assertAnyServerLogIs([('GET', '/test.txt', 202)])

    def test_path(self):
        with self.assertRaises(NotImplementedError):
            self.storage.path('test.txt')

    # Depending on which server is used by get_available_name, this can blow
    # up with "Failed to create" or "Failed to check", so we loosen the check
    # in the log file to the longest common prefix and don't check queries.
    def test_save_dilettante(self):
        self.http_server.override_code = 202
        with self.assertRaises(UnexpectedStatusCode):
            self.storage.save('test.txt', ContentFile(CONTENT))
        self.assertIn("Failed to c", self.get_log())

    def test_size_broken(self):
        if len(self.storage.hosts) > 1:
            return
        self.http_server.override_code = 500
        self.create_file('test.txt', CONTENT)
        with self.assertRaises(HTTPError):
            self.storage.size('test.txt')
        self.assertIn("Failed to get the size", self.get_log())
        self.assertAnyServerLogIs([('HEAD', '/test.txt', 500)])

    def test_size_dilettante(self):
        if len(self.storage.hosts) > 1:
            return
        self.http_server.override_code = 202
        self.create_file('test.txt', CONTENT)
        with self.assertRaises(UnexpectedStatusCode):
            self.storage.size('test.txt')
        self.assertIn("Failed to get the size", self.get_log())
        self.assertAnyServerLogIs([('HEAD', '/test.txt', 202)])


class CommonMiscTestCaseMixin(object):

    # Ensure the storage backends are pickleable, for task queues.

    def test_pickle(self):
        pickled = pickle.dumps(self.storage, pickle.HIGHEST_PROTOCOL)
        unpickled = pickle.loads(pickled)
        self.assertEqual(type(unpickled), type(self.storage))
        self.assertEqual(unpickled.hosts, self.storage.hosts)


class HybridMiscTestCaseMixin(CommonMiscTestCaseMixin):

    # This test case exercises some code paths that are not stressed by the
    # main tests because they handle unexpected situations. For coverage.

    def test_invalid_base_url(self):
        with self.assertRaises(ValueError):
            self.storage_class(base_url='http://example.com/?a_query_does_not_make_sense')
        with self.assertRaises(ValueError):
            self.storage_class(base_url='http://example.com/#a_fragment_does_not_make_sense')

    def test_open_only_for_read(self):
        self.create_file('test.txt', CONTENT)
        self.storage._open('test.txt', 'rb').close()
        with self.assertRaises(IOError):
            self.storage._open('test.txt', 'wb')


class AsyncMiscTestCaseMixin(HybridMiscTestCaseMixin):

    pass


class DistributedMiscTestCaseMixin(CommonMiscTestCaseMixin):

    def test_fatal_exceptions_disabled(self):
        DistributedStorage.fatal_exceptions = False
        try:
            DistributedStorage()
            self.assertIn("You have been warned.", self.get_log())
        finally:
            DistributedStorage.fatal_exceptions = True

    def test_save_over_existing_file(self):
        self.storage._save('test.txt', ContentFile(CONTENT))
        self.assertEqual(self.get_file('test.txt'), CONTENT)
        self.storage._save('test.txt', ContentFile(CONTENT * 2))
        self.assertEqual(self.get_file('test.txt'), CONTENT * 2)
        self.assertIn("PUT on existing file", self.get_log())


class UseDistributedStorageMixin(object):

    storage_class = DistributedStorage
    use_fs = False


class UseHybridStorageMixin(object):

    storage_class = HybridStorage
    use_fs = True


class UseAsyncStorageMixin(object):

    storage_class = AsyncStorage
    use_fs = True

    def sync(self):
        # Wait until only the main thread and the HTTP servers are running.
        while threading.active_count() > 1 + getattr(self, 'num_threads', 0):
            time.sleep(0.001)


class DistributedStorageTestCase(
        UseDistributedStorageMixin, DistributedStorageTestCaseMixin,
        StorageUtilitiesMixin, unittest.TestCase):

    pass


class HybridStorageTestCase(
        UseHybridStorageMixin, HybridStorageTestCaseMixin,
        StorageUtilitiesMixin, unittest.TestCase):

    pass


class AsyncStorageTestCase(
        UseAsyncStorageMixin, AsyncStorageTestCaseMixin,
        StorageUtilitiesMixin, unittest.TestCase):

    pass


class DistributedStorageWithTwoServersTestCase(
        UseDistributedStorageMixin, DistributedStorageTestCaseMixin,
        StorageUtilitiesWithTwoServersMixin, unittest.TestCase):

    pass


class HybridStorageWithTwoServersTestCase(
        UseHybridStorageMixin, HybridStorageTestCaseMixin,
        StorageUtilitiesWithTwoServersMixin, unittest.TestCase):

    pass


class AsyncStorageWithTwoServersTestCase(
        UseAsyncStorageMixin, AsyncStorageTestCaseMixin,
        StorageUtilitiesWithTwoServersMixin, unittest.TestCase):

    pass


class DistributedStorageMiscTestCase(
        UseDistributedStorageMixin, DistributedMiscTestCaseMixin,
        StorageUtilitiesMixin, unittest.TestCase):

    pass


class HybridStorageMiscTestCase(
        UseHybridStorageMixin, HybridMiscTestCaseMixin,
        StorageUtilitiesMixin, unittest.TestCase):

    pass


class AsyncStorageMiscTestCase(
        UseAsyncStorageMixin, AsyncMiscTestCaseMixin,
        StorageUtilitiesMixin, unittest.TestCase):

    pass



"""Default settings. Override them in the settings file of your project.

See the README for more information.
"""

from __future__ import unicode_literals

from django.conf import settings


def get_setting(name):
    name = 'RESTO_%s' % name
    # raise a KeyError if we have no such setting
    default = globals()[name]
    return getattr(settings, name, default)


RESTO_TIMEOUT = 2

RESTO_MEDIA_HOSTS = ()

RESTO_FATAL_EXCEPTIONS = True

RESTO_SHOW_TRACEBACK = False



# coding: utf-8

from __future__ import unicode_literals

import os.path
import threading
try:
    from urllib.request import URLError, urlopen
except ImportError:
    from urllib2 import URLError, urlopen


from django.conf import settings
from django.utils import unittest

from .http_server import TestHttpServer
from .storage import GetRequest, HeadRequest, DeleteRequest, PutRequest


CONTENT = '¡test!'.encode('utf-8')


class HttpServerTestCaseMixin(object):

    host = str('localhost')
    port = 4080
    filename = str('test.txt')
    path = str('/') + filename
    url = str('http://%s:%d%s') % (host, port, path)
    filepath = os.path.join(settings.MEDIA_ROOT, filename)
    num_threads = 1

    def setUp(self):
        super(HttpServerTestCaseMixin, self).setUp()
        self.http_server = TestHttpServer(self.host, self.port)
        self.thread = threading.Thread(target=self.http_server.run)
        self.thread.daemon = True
        self.thread.start()

    def tearDown(self):
        self.http_server.stop()
        self.thread.join()
        self.http_server.server_close()
        super(HttpServerTestCaseMixin, self).tearDown()

    def assertHttpSuccess(self, *args):
        return urlopen(*args).read()            # URLError not raised.

    def assertHTTPErrorCode(self, code, *args):
        with self.assertRaises(URLError) as context:
            urlopen(*args)
        self.assertEqual(context.exception.code, code, 'Expected HTTP %d, '
                'got HTTP %d' % (code, context.exception.code))

    def sync(self):
        pass

    def assertServerLogIs(self, log):
        self.sync()
        self.assertEqual(log, self.http_server.log)

    assertEachServerLogIs = assertServerLogIs

    assertAnyServerLogIs = assertServerLogIs


class ExtraHttpServerTestCaseMixin(object):

    num_threads = 2

    def setUp(self):
        super(ExtraHttpServerTestCaseMixin, self).setUp()
        self.alt_http_server = TestHttpServer(self.host, self.port + 1)
        self.alt_thread = threading.Thread(target=self.alt_http_server.run)
        self.alt_thread.daemon = True
        self.alt_thread.start()
        self.alt_url = str('http://%s:%d/') % (self.host, self.port + 1)

    def tearDown(self):
        self.alt_http_server.stop()
        self.alt_thread.join()
        self.alt_http_server.server_close()
        super(ExtraHttpServerTestCaseMixin, self).tearDown()

    def assertAltServerLogIs(self, log):
        self.sync()
        self.assertEqual(log, self.alt_http_server.log)

    def assertEachServerLogIs(self, log):
        self.sync()
        self.assertServerLogIs(log)
        self.assertAltServerLogIs(log)

    def assertAnyServerLogIs(self, log):
        self.sync()
        return search_merge(log, self.http_server.log, self.alt_http_server.log)


def search_merge(log, log1, log2):
    if log == []:
        return log1 == log2 == []
    if log1 == []:
        return log == log2
    if log2 == []:
        return log == log1
    return (
        (log[0] == log1[0] and search_merge(log[1:], log1[1:], log2))
        or
        (log[0] == log2[0] and search_merge(log[1:], log1, log2[1:]))
    )


class HttpServerTestCase(HttpServerTestCaseMixin, unittest.TestCase):

    def test_get(self):
        self.assertHTTPErrorCode(404, GetRequest(self.url))
        self.http_server.create_file(self.filename, CONTENT)
        body = self.assertHttpSuccess(GetRequest(self.url))
        self.assertEqual(body, CONTENT)
        self.assertServerLogIs([
            ('GET', self.path, 404),
            ('GET', self.path, 200),
        ])

    def test_head(self):
        self.assertHTTPErrorCode(404, HeadRequest(self.url))
        self.http_server.create_file(self.filename, CONTENT)
        body = self.assertHttpSuccess(HeadRequest(self.url))
        self.assertEqual(body, b'')
        self.assertServerLogIs([
            ('HEAD', self.path, 404),
            ('HEAD', self.path, 200),
        ])

    def test_delete(self):
        # delete a non-existing file
        self.assertHTTPErrorCode(404, DeleteRequest(self.url))
        # delete an existing file
        self.http_server.create_file(self.filename, CONTENT)
        body = self.assertHttpSuccess(DeleteRequest(self.url))
        self.assertEqual(body, b'')
        self.assertFalse(self.http_server.has_file(self.filename))
        # attempt to put in read-only mode
        self.http_server.create_file(self.filename, CONTENT)
        self.http_server.readonly = True
        self.assertHTTPErrorCode(403, DeleteRequest(self.url, CONTENT))
        self.assertServerLogIs([
            ('DELETE', self.path, 404),
            ('DELETE', self.path, 204),
            ('DELETE', self.path, 403),
        ])

    def test_put(self):
        # put a non-existing file
        body = self.assertHttpSuccess(PutRequest(self.url, CONTENT))
        self.assertEqual(body, b'')
        self.assertEqual(self.http_server.get_file(self.filename), CONTENT)
        # put an existing file
        body = self.assertHttpSuccess(PutRequest(self.url, CONTENT * 2))
        self.assertEqual(body, b'')
        self.assertEqual(self.http_server.get_file(self.filename), CONTENT * 2)
        # attempt to put in read-only mode
        self.http_server.readonly = True
        self.assertHTTPErrorCode(403, PutRequest(self.url, CONTENT))
        self.assertServerLogIs([
            ('PUT', self.path, 201),
            ('PUT', self.path, 204),
            ('PUT', self.path, 403),
        ])



from __future__ import unicode_literals

import os

DATABASES = {
    'default': {'ENGINE': 'django.db.backends.sqlite3', 'NAME': ':memory:'},
}

DEFAULT_FILE_STORAGE = 'django_resto.storage.HybridStorage'

INSTALLED_APPS = ('django_resto',)

MEDIA_ROOT = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'tests', 'media')

MEDIA_URL = 'http://media.example.com/'

SECRET_KEY = 'Secret key for django-resto tests.'

del os



from __future__ import unicode_literals

import socket
try:                                                        # cover: disable
    from http.server import BaseHTTPRequestHandler, HTTPServer
    from urllib.parse import unquote
    from urllib.request import URLError, urlopen
except ImportError:
    from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
    from urllib import unquote
    from urllib2 import URLError, urlopen


class TestHttpServerRequestHandler(BaseHTTPRequestHandler):

    """Request handler for the test HTTP server.

    Console logging is disabled to avoid spurious output during the tests.
    """

    @property
    def filename(self):
        if isinstance(self.path, bytes):                # Python 2
            return unquote(self.path.lstrip(b'/')).decode('utf-8')
        else:                                           # cover: disable
            return unquote(self.path.lstrip('/'))

    @property
    def content(self):
        return self.rfile.read(int(self.headers['Content-Length']))

    def safe(self, include_content=True):
        try:
            content = self.server.get_file(self.filename)
        except KeyError:
            self.send_error(404)
        else:
            self.send_response(200)
            self.send_header('Content-Length', len(content))
            self.end_headers()
            if include_content:
                self.wfile.write(content)

    def no_content(self, code=204):
        self.send_response(code)
        self.send_header('Content-Length', 0)
        self.end_headers()

    def do_GET(self):
        return self.safe()

    def do_HEAD(self):
        return self.safe(include_content=False)

    def do_PUT(self):
        if self.server.readonly:
            self.send_error(403)
            return
        created = not self.server.has_file(self.filename)
        self.server.create_file(self.filename, self.content)
        self.no_content(201 if created else 204)

    def do_DELETE(self):
        if self.server.readonly:
            self.send_error(403)
            return
        try:
            self.server.delete_file(self.filename)
        except KeyError:
            self.send_error(404)
        else:
            self.no_content()

    def log_request(self, code=None, size=None):
        code = self.server.override_code or code
        self.server.log.append((self.command, self.path, code))

    def log_message(self, *args):
        pass    # disable logging

    def send_response(self, code, message=None):
        BaseHTTPRequestHandler.send_response(self,
            self.server.override_code or code, message)


class TestHttpServer(HTTPServer):

    """Test HTTP server.

    This class provides a basic, in-memory implementation of GET, HEAD, PUT
    and DELETE, as well as a few methods to manage the pseudo-files.

    self.log keeps a record of (method, path, response code) for each query.

    When self.override_code isn't None, this HTTP status code is returned in
    every response.

    When self.readonly is True, PUT and DELETE requests are forbidden.

    self.override_code and self.readonly are used to test invalid behaviors.
    """

    def __init__(self, host=str('localhost'), port=4080):
        self.files = {}
        self.log = []
        self.override_code = None
        self.readonly = False
        self.running = True
        HTTPServer.__init__(self, (host, port), TestHttpServerRequestHandler)

    def has_file(self, name):
        """Test if a file exists on the server."""
        return name in self.files

    def get_file(self, name):
        """Obtain the contents of a file on the server."""
        return self.files[name]

    def create_file(self, name, content):
        """Create a file on the server."""
        self.files[name] = content

    def delete_file(self, name):
        """Delete a file on the server."""
        del self.files[name]

    def run(self):
        """Start the server.

        It will handle requests until it receives a request with a STOP method.
        """
        while self.running:
            self.handle_request()

    def stop(self, timeout=0.1):
        """Stop the server."""
        if self.running:
            self.running = False
            # Make a random query to unblock the main loop.
            try:
                urlopen(str('http://%s:%d/') % self.server_address, timeout=timeout)
            except (URLError, socket.timeout):
                pass
        else:                                               # cover: disable
            print("Warning: stop() called with server wasn't running!")



from __future__ import unicode_literals

import logging
import random
import sys
import threading
try:                                                        # cover: disable
    from urllib.parse import quote, urljoin, urlsplit, urlunsplit
    from urllib.request import HTTPError, Request, URLError, urlopen
except ImportError:
    from urllib import quote
    from urllib2 import HTTPError, Request, URLError, urlopen
    from urlparse import urljoin, urlsplit, urlunsplit

from django.conf import settings
from django.core.files.base import ContentFile
from django.core.files.storage import Storage, FileSystemStorage
from django.utils.encoding import filepath_to_uri

from .settings import get_setting


logger = logging.getLogger(__name__)


class UnexpectedStatusCode(HTTPError):

    """Exception raised when a server returns an unexpected status code.

    Only the most common codes (200, 201, 204, 404) are interpreted by
    DefaultTransport. If it receives another code, it will raise this
    exception. It won't try to interpret the class of the status code.

    For instance, "202 Accepted" indicates a successful request, but it
    breaks our expectation that the upload is synchronous. So we'd better
    raise an exception. Other uncommon codes pose similar problems.
    """

    def __init__(self, resp):
        super(UnexpectedStatusCode, self).__init__(
            resp.url, resp.code, resp.msg, resp.headers, resp.fp)


class GetRequest(Request):
    """HTTP GET request."""
    # This adds nothing to urllib, but it's there for consistency.
    def get_method(self):
        return str('GET')


class HeadRequest(Request):
    """HTTP HEAD request."""
    def get_method(self):
        return str('HEAD')


class DeleteRequest(Request):
    """HTTP DELETE request."""
    def get_method(self):
        return str('DELETE')


class PutRequest(Request):
    """HTTP PUT request."""
    def get_method(self):
        return str('PUT')


class DefaultTransport(object):
    """Transport to read and write files over HTTP.

    This transport expects that the target HTTP hosts implements the GET,
    HEAD, PUT and DELETE methods according to RFC2616.
    """

    timeout = get_setting('TIMEOUT')

    def __init__(self, base_url):
        scheme, netloc, path, query, fragment = urlsplit(str(base_url))
        if query or fragment:
            raise ValueError('base_url may not contain a query or fragment.')
        self.scheme = scheme or str('http')
        self.path = path or str('/')

    ### Hooks for custom transports

    def _get_url(self, host, name):
        """Return the full URL for a file on a given host."""
        path = self.path + quote(name.encode('utf-8'))
        return urlunsplit((self.scheme, str(host), path, str(''), str('')))

    def _http_request(self, request):
        """Return a response object for a given request."""
        return urlopen(request, timeout=self.timeout)

    ### Wrappers around HTTP methods

    def content(self, host, name):
        """Get the content of a file as a string.

        URLError will be raised if something goes wrong.
        """
        url = self._get_url(host, name)
        resp = self._http_request(GetRequest(url))
        if resp.code != 200:
            raise UnexpectedStatusCode(resp)
        length = resp.info().get('Content-Length')
        if length is None:                                  # cover: disable
            return resp.read()
        else:
            return resp.read(int(length))

    def exists(self, host, name):
        """Check if a file exists.

        Return True if the file exists, False if it doesn't.

        URLError will be raised if something goes wrong.
        """
        url = self._get_url(host, name)
        try:
            resp = self._http_request(HeadRequest(url))
            if resp.code != 200:
                raise UnexpectedStatusCode(resp)
            return True
        except HTTPError as e:
            if e.code not in (404, 410):
                raise
            return False

    def size(self, host, name):
        """Check the size of a file.

        This method relies on the Content-Length header.

        URLError will be raised if something goes wrong.
        """
        url = self._get_url(host, name)
        resp = self._http_request(HeadRequest(url))
        if resp.code != 200:
            raise UnexpectedStatusCode(resp)
        length = resp.info().get('Content-Length')
        if length is None:
            raise NotImplementedError("The HTTP server did not provide a"
                    "content length for %r." % resp.geturl())
        return int(length)

    def create(self, host, name, content):
        """Create or update a file.

        Return True if the file existed, False if it did not.

        URLError will be raised if something goes wrong.
        """
        url = self._get_url(host, name)
        resp = self._http_request(PutRequest(url, content))
        if resp.code == 201:
            return False
        elif resp.code == 204:
            logger.warning("PUT on existing file %s on %s.", name, host)
            return True
        else:
            raise UnexpectedStatusCode(resp)

    def delete(self, host, name):
        """Delete a file.

        Return True if the file existed, False if it did not.

        URLError will be raised if something goes wrong.
        """
        url = self._get_url(host, name)
        try:
            resp = self._http_request(DeleteRequest(url))
            if resp.code not in (200, 204):
                raise UnexpectedStatusCode(resp)
            return True
        except HTTPError as e:
            if e.code not in (404, 410):
                raise
            logger.warning("DELETE on missing file %s on %s.", name, host)
            return False


class DistributedStorageMixin(object):

    """Mixin for storage backends that distribute files on several servers."""

    fatal_exceptions = get_setting('FATAL_EXCEPTIONS')
    show_traceback = get_setting('SHOW_TRACEBACK')

    def __init__(self, hosts=None, base_url=None, transport=DefaultTransport):
        if hosts is None:                                   # cover: disable
            hosts = get_setting('MEDIA_HOSTS')
        self.hosts = hosts
        if base_url is None:                                # cover: disable
            base_url = settings.MEDIA_URL
        self.base_url = base_url
        self.transport = transport(base_url=base_url)

    def execute(self, func, url, *args, **kwargs):
        """Run an action over several hosts in parallel."""
        exceptions = {}

        def execute_inner(host):
            try:
                func(host, url, *args, **kwargs)
            except Exception:
                exceptions[host] = sys.exc_info()

        if len(self.hosts) == 1:
            execute_inner(self.hosts[0])
        else:
            threads = [threading.Thread(target=execute_inner, args=(host,))
                    for host in self.hosts]
            for thread in threads:
                thread.start()
            for thread in threads:
                thread.join()

        for host, exc_info in exceptions.items():
            action = func.__name__
            logger.error("Failed to %s %s on %s.", action, url, host,
                    exc_info=exc_info if self.show_traceback else None)

        if exceptions and self.fatal_exceptions:
            # Let's raise a random exception, we've logged them all anyway
            raise exceptions.popitem()[1][1]


class DistributedStorage(DistributedStorageMixin, Storage):

    """Backend that stores files remotely over HTTP."""

    def __init__(self, hosts=None, base_url=None):
        DistributedStorageMixin.__init__(self, hosts, base_url)
        if not self.fatal_exceptions:
            logger.warning("You're using the DistributedStorage backend with "
                    "RESTO_FATAL_EXCEPTIONS = %r.", self.fatal_exceptions)
            logger.warning("This is prone to data-loss problems, and I won't "
                    "take any responsibility in what happens from now on.")
            logger.warning("You have been warned.")
        Storage.__init__(self)

    ### Hooks for custom storage objects

    def _open(self, name, mode='rb'):
        # Allowing writes would be doable, if we distribute the file to the
        # media servers when it's closed. Let's forbid it for now.
        if mode != 'rb':                                    # cover: disable
            raise IOError('Unsupported mode %r, use %r.' % (mode, 'rb'))
        host = random.choice(self.hosts)
        try:
            return ContentFile(self.transport.content(host, name))
        except URLError:
            logger.error("Failed to download %s from %s.", name, host,
                    exc_info=self.show_traceback)
            raise

    def _save(self, name, content):
        # It's hard to avoid buffering the whole file in memory,
        # because different threads will read it simultaneously.
        self.execute(self.transport.create, name, content.read())
        return name

    ### Mandatory methods

    # The implementations of get_valid_name, get_available_name, and path
    # in Storage are OK for DistributedStorage.

    def delete(self, name):
        self.execute(self.transport.delete, name)

    def exists(self, name):
        host = random.choice(self.hosts)
        try:
            return self.transport.exists(host, name)
        except URLError:
            logger.error("Failed to check if %s exists on %s.", name, host,
                    exc_info=self.show_traceback)
            raise

    # It is not possible to implement listdir in pure HTTP. It could
    # be done with WebDAV.

    def size(self, name):
        host = random.choice(self.hosts)
        try:
            return self.transport.size(host, name)
        except URLError:
            logger.error("Failed to get the size of %s from %s.", name, host,
                    exc_info=self.show_traceback)
            raise

    def url(self, name):
        return urljoin(self.base_url, filepath_to_uri(name))


class HybridStorage(DistributedStorageMixin, FileSystemStorage):

    """Backend that stores files both locally and remotely over HTTP."""

    def __init__(self, hosts=None, base_url=None, location=None):
        DistributedStorageMixin.__init__(self, hosts, base_url)
        FileSystemStorage.__init__(self, location, base_url)

    # Read operations can be done with FileSystemStorage. Write operations
    # must be done with FileSystemStorage and DistributedStorageMixin, in
    # this order.

    ### Hooks for custom storage objects

    def _open(self, name, mode='rb'):
        # Writing is forbidden, see DistributedStorage._open.
        if mode != 'rb':                                    # cover: disable
            raise IOError('Unsupported mode %r, use %r.' % (mode, 'rb'))
        return FileSystemStorage._open(self, name, mode)

    def _save(self, name, content):
        name = FileSystemStorage._save(self, name, content)
        # After this line, we will assume that 'name' is available on the
        # media servers. This could be wrong if a delete for this file name
        # failed at some point in the past.
        self.execute(self.upload, name)
        return name

    # Make this a separate method so it can be passed to a task queue.
    def upload(self, host, name):
        with self.open(name) as handle:
            self.transport.create(host, name, handle.read())

    ### Mandatory methods

    # The implementations of get_valid_name, get_available_name, path, exists,
    # listdir, size, and url in FileSystemStorage are OK for HybridStorage.

    def delete(self, name):
        FileSystemStorage.delete(self, name)
        self.execute(self.transport.delete, name)


class AsyncStorage(HybridStorage):

    """Backend that stores files both locally and remotely over HTTP."""

    def execute(self, func, url, *args, **kwargs):
        """Run an action over several hosts asynchronously."""
        for host in self.hosts:
            self.execute_one(func, host, url, *args, **kwargs)

    def execute_one(self, func, host, url, *args, **kwargs):
        """Run a single action asynchronously."""
        def execute_inner():
            try:
                func(host, url, *args, **kwargs)
            except Exception:
                action = func.__name__
                logger.error("Failed to %s %s on %s.", action, url, host,
                        exc_info=self.show_traceback)
        return threading.Thread(target=execute_inner).start()



from __future__ import unicode_literals

from django.utils import unittest

from .settings import get_setting


class SettingsTestCase(unittest.TestCase):

    def test_get_existing_setting(self):
        self.assertTrue(get_setting('TIMEOUT') > 0)

    def test_get_non_existing_setting(self):
        with self.assertRaises(KeyError):
            get_setting('TIME-OUT')



# coding: utf-8
from __future__ import unicode_literals

from django.utils import unittest

from .test_storage import StorageUtilitiesMixin, UseDistributedStorageMixin


class RegressionTestCase(
    StorageUtilitiesMixin, UseDistributedStorageMixin, unittest.TestCase):

    def test_non_ascii_file_name(self):
        self.create_file('café.txt', b'caffeine')
        self.assertTrue(self.storage.exists('café.txt'))
        self.delete_file('café.txt')
        self.assertFalse(self.storage.exists('café.txt'))
